/*==============================================================================*/
/*  Copyright (C) 2009-2021, Heartland Data inc. All Rights Reserved.           */
/*                                                                              */
/*  Title  :   CAN Extended Driver                                              */
/*  EventID:   Insert EventID Test Point                                        */
/*  AppVer :   DT10 Ver10.XX`@                                                */
/*  FileID :   11C1                                                             */
/*  Version:   1.1                                                              */
/*  Author :   HLDC                                                             */
/*==============================================================================*/

#include "Common.h"
#include "MCP2515.h"

#define MAX_FILE 511
#define MAX_FILE_BIT ((MAX_FILE+31)/32)
#define MAX_MODULE 255
#define	TP_TIME_OUT	(10*1000/10)	// 10sec
#define	MAX_MESSAGE	15

static unsigned int DtSecurityCode[4] = {
	0x8197C80F,
	0x7B211334,
	0x9754BD69,
	0x19F32567
};

static unsigned int TPFilter;
static unsigned int SrcFilter[MAX_FILE_BIT];
static unsigned char can_src_base_addr[MAX_FILE];
static unsigned char can_mdl_no;
static unsigned char can_auth_phase;
static unsigned char can_auth_ok;
static unsigned char can_auth_start;
static unsigned char can_data_send_ok;
static unsigned char can_data_buff_ok;
static unsigned char can_src_fltr;
static unsigned char can_data_type;
static unsigned int can_timer;
static void (*can_tp_busout_func)(unsigned int addr, unsigned int dat) = NULL;
static void (*can_tp_busvar_func)(unsigned int addr, unsigned int dat, void *p, unsigned int size) = NULL;
static int (*can_tp_senddata_func)(unsigned char *buf) = NULL;

#define MAX_LOG	(8*MAX_MESSAGE)
static int logWriteLen = MAX_LOG;
static int logReadLen = 0;
static int logWritePos = 0;
static int logReadPos = 0;
static unsigned char logBuff[MAX_LOG];

//==============================================================================
//	Macro:	Test Point ID
//==============================================================================
#define		DT_VARIABLE_BIT			0x02
#define		DT_TIME_INFO_BIT		0x04
#define		DT_EVTTRG_BIT			0x08

//------------------------------------------------------------------------------
//  Macro:	DT_ADD_TIME_INFO
//	Desc:	Please set 1 if use TARGET SYSTEM TIME information with Test Point.
//------------------------------------------------------------------------------
#define DT_ADD_TIME_INFO 1

//------------------------------------------------------------------------------
//  Macro:	DT_UINT
//	Desc:	Please chage Test Point argument type for DT10 Project setting.
//------------------------------------------------------------------------------
#define	DT_UINT unsigned int

//------------------------------------------------------------------------------
//	Macro:	DT_INLINE
//	Desc: 	Please use "static" instead of "inline" if "inline" cannot be used.
//------------------------------------------------------------------------------
//#define	DT_INLINE	inline
#define	DT_INLINE	static

//------------------------------------------------------------------------------
//	Func:	_TP_BusPortInit
//	Desc: 	Please describe the code to initializes ports.
//------------------------------------------------------------------------------
void _TP_BusPortInit(void)
{
	MCP2515Config();

	SysCtlPeripheralEnable( SYSCTL_PERIPH_GPIOD );
	GPIOPinTypeGPIOOutput( GPIO_PORTD_BASE, GPIO_PIN_3 );		/*	CS	*/
	GPIOPinWrite( GPIO_PORTD_BASE, GPIO_PIN_3, 0);
	GPIOPinTypeGPIOOutput( GPIO_PORTD_BASE, GPIO_PIN_2 );		/*	CLK	*/
	GPIOPinWrite( GPIO_PORTD_BASE, GPIO_PIN_2, 0);

}

//------------------------------------------------------------------------------
//	Func:	enterCritical
//	Desc:	Please describe the code to enter Critical Section.
//------------------------------------------------------------------------------
DT_INLINE void enterCritical(void)
{
//	ex1)
	taskENTER_CRITICAL();
//	ex2)
//	DI();
}

//------------------------------------------------------------------------------
//	Func:	exitCritical
//	Desc:	Please describe the code to exit Critical Section.
//------------------------------------------------------------------------------
DT_INLINE void exitCritical(void)
{
//	ex1)
	taskEXIT_CRITICAL();
//	ex2)
//	EI();
}

#if DT_ADD_TIME_INFO
//------------------------------------------------------------------------------
//	Func:	_TP_GetTargetTime(void)
//	Desc:	
//------------------------------------------------------------------------------
DT_INLINE unsigned int _TP_GetTargetTime(void)
{
	return GetSystemTimer();	
}
#endif

//------------------------------------------------------------------------------
//	Func: _TP_FilterCheck
//	Desc: check Test Point and Source Code filter information
//------------------------------------------------------------------------------
static int _TP_FilterCheck( DT_UINT addr, DT_UINT dat )
{
	int key;

	if( (TPFilter & dat) == 0 ){
		return 0;
	}
	if( can_src_fltr ){
		key = addr >> 4;
		if( (SrcFilter[key>>5] & (1 << (key&0x1f))) == 0 ){
			return 0;
		}
	}
	if( can_mdl_no ){
		if( (dat >> 24) != can_mdl_no ){
			return 0;
		}
	}
	return 1;
}

//------------------------------------------------------------------------------
//	Func: _TP_LogInit
//	Desc: initialize log data buffer
//------------------------------------------------------------------------------
static void _TP_LogInit(void)
{
	int i;
	logWriteLen = MAX_LOG;
	logReadLen = 0;
	logWritePos = 0;
	logReadPos = 0;
	for( i = 0; i < MAX_LOG; i++ ){
		logBuff[i] = 0;
	}
}

//------------------------------------------------------------------------------
//	Func: _TP_LogStore
//	Desc: store log data to the buffer
//------------------------------------------------------------------------------
static int _TP_LogStore( unsigned char *log, int size )
{
	int	i;
	if( logWriteLen < size ){
		can_data_buff_ok = 0;
		can_data_send_ok = 1;
		return 0;
	}
	enterCritical();
#if 1
	memcpy( &logBuff[logWritePos], log, size );
#else	
	for( i = 0; i < size; i++ ){
		logBuff[logWritePos + i] = *(log+i);		
	}
#endif	
	logWritePos += size;
	logReadLen += size;
	logWriteLen -= size;
	if( logWriteLen == 0 ){
		can_data_buff_ok = 0;
		can_data_send_ok = 1;
	}
	exitCritical();
	return size;
}

//------------------------------------------------------------------------------
//	Func: _TP_BusOut_Normal
//	Desc: Called by Test Point
//------------------------------------------------------------------------------
static void _TP_BusOut_Normal( DT_UINT addr, DT_UINT dat )
{
	unsigned char buf[4];
	buf[0] = dat >> 8;
	buf[1] = dat;
	buf[2] = addr >> 8;
	buf[3] = addr;
	_TP_LogStore(buf, 4);
}

//------------------------------------------------------------------------------
//	Func: _TP_BusOut_Time
//	Desc: Called by Test Point
//------------------------------------------------------------------------------
static void _TP_BusOut_Time( DT_UINT addr, DT_UINT dat )
{
	unsigned char buf[8];
	unsigned int time = _TP_GetTargetTime();
	buf[0] = dat >> 8;
	buf[1] = dat;
	buf[2] = addr >> 8;
	buf[3] = addr | 0x04;
	buf[4] = time;
	buf[5] = time >> 8;
	buf[6] = time >> 16;
	buf[7] = time >> 24;
	_TP_LogStore(buf, 8);
}

//------------------------------------------------------------------------------
//	Func: _TP_BusOut_Renumber
//	Desc: Called by Test Point
//------------------------------------------------------------------------------
static void _TP_BusOut_Renumber( DT_UINT addr, DT_UINT dat )
{
	unsigned char buf[2];
	addr >>= 4;
	if( addr > MAX_FILE ){
		addr = 0;
	}
	buf[0] = can_src_base_addr[addr] << 4;
	buf[0] |= (dat >> 8) & 0x0f;
	buf[1] = dat;
	_TP_LogStore(buf, 2);
}
	
//------------------------------------------------------------------------------
//	Func: _TP_MemoryOutput_Normal
//	Desc: Called by Variable Test Point
//------------------------------------------------------------------------------
static void _TP_MemoryOutput_Normal( DT_UINT addr, DT_UINT dat, void *value, DT_UINT size )
{
	unsigned char buf[16];
	unsigned char *p = value;
	int i;
	
	if( size >= 4 ) size = 4;

	buf[0] = (dat >> 8) & 0x0f;
	buf[0]|= size << 4;
	buf[1] = dat;
	buf[2] = addr >> 8;
	buf[3] = addr | DT_VARIABLE_BIT;
	for( i = 0; i < 4; i++ ){
		if( i < size ){
			buf[4+i] = *(p+i);
		}
		else{
			buf[4+i] = 0;
		}
	}
	_TP_LogStore(buf, 8);
}

//------------------------------------------------------------------------------
//	Func: _TP_MemoryOutput_Time
//	Desc: Called by Variable Test Point
//------------------------------------------------------------------------------
static void _TP_MemoryOutput_Time( DT_UINT addr, DT_UINT dat, void *value, DT_UINT size )
{
	unsigned char buf[16];
	unsigned char *p = value;
	unsigned int time = _TP_GetTargetTime();
	int i;
	
	if( size >= 8 ) size = 8;

	buf[0] = (dat >> 8) & 0x0f;
	buf[0]|= size << 4;
	buf[1] = dat;
	buf[2] = addr >> 8;
	buf[3] = addr | DT_VARIABLE_BIT | 0x04;
	buf[4] = time;
	buf[5] = time >> 8;
	buf[6] = time >> 16;
	buf[7] = time >> 24;
	for( i = 0; i < 8; i++ ){
		if( i < size ){
			buf[8+i] = *(p+i);
		}
		else{
			buf[8+i] = 0;
		}
	}
	_TP_LogStore(buf, 16);
}

//------------------------------------------------------------------------------
//	Func: _TP_MemoryOutput_Renumber
//	Desc: Called by Variable Test Point
//------------------------------------------------------------------------------
static void _TP_MemoryOutput_Renumber( DT_UINT addr, DT_UINT dat, void *value, DT_UINT size )
{
}

//------------------------------------------------------------------------------
//	Func: _TP_CanDataSend_Normal
//	Desc: to be called in every 10msec
//------------------------------------------------------------------------------
static int _TP_CanSendData_Normal(unsigned char *buf)
{
	int i;
	if( logBuff[logReadPos+7] & 0x02){	// Variable Test Point?
		if( logReadLen < 4 ) return 0;
		for( i = 0; i < 4; i++ ){
			buf[i] = logBuff[logReadPos+i];
		}
		for( i = 4; i < 8; i++ ){
			buf[i] = 0x00;
		} 
		logReadPos += 4;
		logReadLen -= 4;
		logWriteLen += 4;
		return 4;
	}
	else{
		if( logReadLen < 8 ) return 0;
		for( i = 0; i < 8; i++ ){
			buf[i] = logBuff[logReadPos+i];
		}
		logReadPos += 8;
		logReadLen -= 8;
		logWriteLen += 8;
		return 8;
	}
}

//------------------------------------------------------------------------------
//	Func: _TP_CanDataSend_Time
//	Desc: to be called in every 10msec
//------------------------------------------------------------------------------
static int _TP_CanSendData_Time(unsigned char *buf)
{
	int i;
	if( logReadLen < 8 ) return 0;
	for( i = 0; i < 8; i++ ){
		buf[i] = logBuff[logReadPos+i];
	}
	logReadPos += 8;
	logReadLen -= 8;
	logWriteLen += 8;
	return 8;
}

//------------------------------------------------------------------------------
//	Func: _TP_CanDataSend_Renumber
//	Desc: to be called in every 10msec
//------------------------------------------------------------------------------
static int _TP_CanSendData_Renumber(unsigned char *buf)
{
	int i;
	if( logReadLen < 2 ) return 0;
	for( i = 0; i < 8; i++ ){
		if( i >= logReadLen ){
			buf[i] = 0x00;
		}
		else{
			buf[i] = logBuff[logReadPos+i];
		}
	}
	logReadPos += 8;
	logReadLen -= 8;
	logWriteLen += 8;
	return 8;
}

//------------------------------------------------------------------------------
//	Func: _TP_BusOut
//	Desc: Called by Test Point
//------------------------------------------------------------------------------
void _TP_BusOut( DT_UINT addr, DT_UINT dat )
{
	if( can_data_buff_ok != 0 ){
		if( _TP_FilterCheck( addr, dat ) ){
			(*can_tp_busout_func)(addr, dat);
			can_timer = TP_TIME_OUT;
		}
	}
}
//------------------------------------------------------------------------------
//	Func: _TP_MemoryOutput
//	Desc: Called by Variable Test Point
//------------------------------------------------------------------------------
void _TP_MemoryOutput( DT_UINT addr, DT_UINT dat, void *value, DT_UINT size )
{
	if( can_tp_busvar_func == NULL ) return;
	if( _TP_FilterCheck( addr, dat ) ){
		(*can_tp_busvar_func)(addr, dat, value, size);
		can_timer = TP_TIME_OUT;
	}
}

//------------------------------------------------------------------------------
//	Func: _TP_EventTrigger
//	Desc: Called by Event Trigger Test Point
//------------------------------------------------------------------------------
void _TP_EventTrigger( DT_UINT addr, DT_UINT dat )
{
}

//------------------------------------------------------------------------------
//	Func: _TP_CanInit
//	Desc: Initialize CAN data
//------------------------------------------------------------------------------
void _TP_CanInit(void)
{
	int i;
	// Filter Data Clear
	for( i = 0; i < MAX_FILE_BIT; i++ ){
		SrcFilter[i] = 0x00000000;
	}
	for( i = 0; i < MAX_FILE; i++ ){
		can_src_base_addr[i] = 0x00;
	}
	TPFilter = 0x00000000;
	// Phase Init
	can_auth_phase = 0;
	can_auth_ok = 0;
	can_auth_start = 0;
	can_data_send_ok = 0;
	can_data_buff_ok = 0;
	can_src_fltr = 0;
	can_mdl_no = 0;
	// Ring Buffer Init
	logWriteLen = MAX_LOG;
	logReadLen = 0;
	logWritePos = 0;
	logReadPos = 0;
	can_tp_busout_func = NULL;
	can_tp_busvar_func = NULL;
	can_tp_senddata_func = NULL;
	can_timer = 0;
}

//------------------------------------------------------------------------------
//	Func: _TP_CanSendCtrl
//	Desc: to be called in every 10msec
//------------------------------------------------------------------------------
void _TP_CanSendCtrl(void)
{
	unsigned char buf[8];
	int size, i;

	if( can_tp_senddata_func == NULL ){
		return;
	}
	if( can_timer != 0 ){
		--can_timer;
		if( can_timer == 0 ){
			can_data_buff_ok = 0;
			can_data_send_ok = 1;
		}
	}
	if( can_data_send_ok == 0 ){
		return;
	}
	do{
		enterCritical();
		size = (*can_tp_senddata_func)(buf);
		exitCritical();
		if( size == 0 ){
			can_data_send_ok = 0;	// Start sending Can Data
			for( i = 0; i < MAX_FILE_BIT; i++ ){
				SrcFilter[i] = 0x00000000;
			}
			for( i = 0; i < MAX_FILE; i++ ){
				can_src_base_addr[i] = 0x00;
			}
			buf[0] = 0x00;
			buf[1] = 0x00;
			buf[2] = 0x00;
			buf[3] = 0x00;
			buf[4] = 0x00;		// f[^擾ʒm
			buf[5] = 0x00;
			buf[6] = 0x00;
			buf[7] = 0x00;
		}
		MCP2515DataSend( buf, 8 );
	}while( size != 0 );
}

//------------------------------------------------------------------------------
//	Func: _TP_CanRecvCtrl
//	Desc: to be called in every 2msec
//------------------------------------------------------------------------------
void _TP_CanRecvCtrl(void)
{
	unsigned char buf[8];
	unsigned int temp, key;
	int i;

	/*** Can Data Check and Receive ***/
	if( MCP2515DataRecv( buf, 8 ) == 0 ){
		return;
	}

	/*** Command Decode ***/
	if( buf[0] != 'D' || buf[1] != 'T' ){
		return;
	}
	switch( buf[2] ){	// Command
	case 0x01:	// Authentication Start Command
		_TP_CanInit();
		can_auth_start = 1;
		MCP2515DataSend( buf, 8 );
		break;
	case 0x02:	// Authentication Code Command
		if( can_auth_start ){
			temp = buf[4];
			temp += buf[5] << 8;
			temp += buf[6] << 16;
			temp += buf[7] << 24;
			if( buf[3] == can_auth_phase &&
				temp == DtSecurityCode[can_auth_phase] ){
//				MCP2515DataSend( buf, 8 );
				++can_auth_phase;
				if( can_auth_phase == 4 ){
					can_auth_ok = 1;
					can_auth_start = 0;
				}
			}
		}
		break;
	case 0x10:	// TP Filter Setting Command
		if( can_auth_ok ){
			TPFilter = (buf[4] & 0x3f) << 16;
//			MCP2515DataSend( buf, 8 );
		}
		break;
	case 0x20:	// Module Filter Setting Command
		if( can_auth_ok ){
			can_data_type = buf[4];
			switch( can_data_type ){
			case 0x00:
				can_tp_busout_func = _TP_BusOut_Normal;
				can_tp_busvar_func = _TP_MemoryOutput_Normal;
				can_tp_senddata_func = _TP_CanSendData_Normal;
				break;
			case 0x01:
				can_tp_busout_func = _TP_BusOut_Time;
				can_tp_busvar_func = _TP_MemoryOutput_Time;
				can_tp_senddata_func = _TP_CanSendData_Time;
				break;
			case 0x02:
				can_tp_busout_func = _TP_BusOut_Renumber;
				can_tp_busvar_func = _TP_MemoryOutput_Renumber;
				can_tp_senddata_func = _TP_CanSendData_Renumber;
				break;
			}
//			MCP2515DataSend( buf, 8 );
		}
		break;
	case 0x30:	// Source Base Addres Renumber Command
		if( can_auth_ok ){
			temp = buf[3];
			for( i = 0; i < 4; i += 2 ){
				key = buf[4+i] + (buf[4+i+1] << 8);
				if( key != 0 ){
					can_src_base_addr[key] = temp++;
					SrcFilter[key>>5] |= 1 << (key&0x1f);
				}
			}
			can_src_fltr = 1;
//			MCP2515DataSend( buf, 8 );
		}
		break;
	case 0x40:	// Start
		if( can_auth_ok ){
			if( buf[3] & 0x01 ){
//				can_data_send_ok = 1;	// Start sending Can Data 
			}
			if( buf[3] & 0x02 ){
				can_data_buff_ok = 1;	// Start buffering TP
			}
			can_mdl_no = buf[4];		// Module No.
//			MCP2515DataSend( buf, 8 );
			can_timer = TP_TIME_OUT;
			_TP_LogInit();
		}
		break;
	case 0x41:	// Stop
		can_data_send_ok = 0;
		can_data_buff_ok = 0;
		break;
	default:
		break;
	}
}
// =============================================================================
